using Distributions
using DataFrames
using Roots
#using Compat

include("GSM.jl")
include("Sampler.jl")

dlogP(s::Sampler) = begin
  z = abs(s.u[1])
  u = s.u[2:end]
  back_pred_error = (s.dist.Aᵀx - z*s.dist.AᵀA*u)/s.dist.σₓ²
  dzu = zeros(length(s.u))
  dzu[1] = sign(s.u[1])*(dot(u, back_pred_error) - z)
  dzu[2:end] = z*back_pred_error - s.dist.invC*u
  dzu
end


step(s::Sampler) = begin
	dtdlogPs = s.dt*dlogP(s)
	common_input = s.dt*s.M*(s.u - s.v) #invM goes here.
  iτl = 1/s.τl
  iτh = 1/s.τh
  #if s.τl == 10.
	s.u   += (iτh - iτl)*common_input + iτl*dtdlogPs + sqrt(2*iτl*s.dt).*randn(dim)
	s.v   += (iτh + iτl)*common_input - iτh*dtdlogPs + sqrt(2*iτl*s.dt).*randn(dim)
  u = s.u[2:end]
  z = s.u[1]
  diff = s.dist.Aᵀx - z*s.dist.AᵀA * u
  s.L = ((s.τl == s.τh) ? 0. : dot(s.v-s.u, s.M * (s.v-s.u))/2) + dot(u, s.dist.AᵀA * u)/(2*(1-s.dist.σₓ²)) + dot(diff, s.dist.AᵀA \ diff)/(2s.dist.σₓ²) + z*z/2
  #    s.u   += iτl*dtdlogPs + sqrt(2*iτl*s.dt).*randn(dim)
  #else
  #    s.u   += (iτh - iτl)*common_input + (iτl+iτh)*dtdlogPs + sqrt(2*iτl*s.dt).*randn(dim)
  #    s.v   += (iτh + iτl)*common_input + sqrt(2*iτl*s.dt).*randn(dim)
  #    (s.u, s.v)
  #end
end

#step_inputs(s::Sampler) = begin
#	dtdlogPs = s.dt*dlogP(s)
#  iτl = 1/s.τl
#  iτh = 1/s.τh
#	common_input = s.dt*s.invM*(s.B*s.u - s.v) #invM goes here.
#  u_input = s.dt*s.invM*s.B*s.u
#  v_input = s.dt*s.invM*s.v
#  uu_input = iτh*(s.B' - s.τh/s.τl*eye(length(s.u)))*u_input
#  uv_input = iτh*(s.B' - s.τh/s.τl*eye(length(s.u)))*v_input
#	s.u   += uu_input-uv_input + iτl*dtdlogPs + sqrt(2*iτl*s.dt).*randn(dim)
#  #s.u[1] = s.dist.z
#	s.v   += iτh*(eye(length(s.u)) + s.τh/s.τl*s.B')*common_input - iτh*dtdlogPs + sqrt(2*iτl*s.dt).*randn(dim)
#  (u_input, v_input)
#end

step_inputs = step
#step = step_inputs

import Base.trace
trace(s::Sampler, Tms::Int, z2, interval=1.) = begin
  time_steps_per_interval = round(Int, interval/s.dt)
  #Burn in 100 ms
  for i = 1:round(Int, 100/s.dt)
    step(s)
  end
  us = zeros(round(Int, 2*Tms/interval), length(s.u))
  for t = 1:round(Int, Tms/interval)
    for j = 1:time_steps_per_interval
      step(s)
    end
	  us[t, :] = s.u
  end
  if z2 != nothing
    s.regen(s.dist, z2)
  end
  for t = 1:round(Int, Tms/interval)
    for j = 1:time_steps_per_interval
      step(s)
    end
	  us[round(Int, Tms/interval)+t, :] = s.u
  end
  us
end

import Base.trace
trace_uvs(s::Sampler, Tms::Int, z2::Number) = begin
  time_steps_per_ms = round(Int, 1/s.dt)#/2)
  T = Tms * time_steps_per_ms
  #Burn in 100 ms
  ts = (1:T)/time_steps_per_ms
  us = zeros(T, length(s.u))
  vs = zeros(T, length(s.u))
  Ls = zeros(T)
  s.regen(s.dist, z2)
  for i = 1:(100*time_steps_per_ms)
    step(s)
  end
  for t = 1:T
    step(s)
	  us[t, :] = s.u
	  vs[t, :] = s.v
	  Ls[t] = s.L
  end
  (ts, us, vs, Ls)
end

import Base.trace
trace_vs(s::Sampler, Tms::Int, z2::Number) = begin
  time_steps_per_ms = round(Int, 1/s.dt)#/2)
  #Burn in 100 ms
  for i = 1:(100*time_steps_per_ms)
    step(s)
  end
  us = zeros(2*Tms, length(s.u))
  vs = zeros(2*Tms, length(s.u))
  for t = 1:Tms
    for j = 1:time_steps_per_ms
      step_inputs(s)
    end
	  us[t, :] = s.u
	  vs[t, :] = s.v
  end
  s.regen(s.dist, z2)
  for t = 1:Tms
    for j = 1:time_steps_per_ms
      step_inputs(s)
    end
	  us[Tms+t, :] = s.u
	  vs[Tms+t, :] = s.v
  end
  (us, vs)
end

traces(;zs::Vector=[1.], 
        reps::Integer=1, 
        Tms::Integer=1000, 
        different_models::Bool=true,
        regen::Function=regen_contrast,
        n::Int=15,
        τl=150.,
        τh::Float64=10.) = begin
  s = Sampler(GSM(), regen=regen, τh=τh, τl=float(τl))
  uss = Vector{Matrix{Float64}}[Array(Matrix{Float64}, reps) for i in 1:length(zs)]
  for z_index = 1:length(zs)
    for rep = 1:reps
      s.regen(s.dist, 0)
      uss[z_index][rep] = trace(different_models ? Sampler(GSM(), regen=regen, τh=τh, τl=float(τl)) : s, Tms, zs[z_index])
    end
  end
  (uss, zs)
end

const gsm = GSM()
z_comp = 1.
import Base.mean
mean(g::GSM) = begin
    regen_contrast(g, z_comp)
    total_w = 0.
    total_m = zeros(n)
    C = inv(g.invC)
    zs = 0.01:0.01:20.
#    zs = [g.z_true]
    ws = zeros(length(zs))
    for i = 1:length(zs) 
        z = zs[i]
#        S = inv(g.invC + z*z*g.AᵀA/g.σₓ²)
        S = inv(g.invC + z*z*g.AᵀA/g.σₓ²)
        m = (z/g.σₓ²) * S * g.Aᵀx
#        w = 1.
        w = pdf(Normal(), z)*
            #Woodbury
            exp(dot(g.Aᵀx, (g.invC/(z*z) + g.AᵀA/g.σₓ²) \ g.Aᵀx)/(2*g.σₓ²^2)-
            #Generalisation of Sylvester matrix identity
                logdet(g.invC + z*z/g.σₓ²*g.AᵀA)/2)
        ws[i] = w
        total_w += w
        total_m += w*m
    end
    regen_contrast(g, 0.)
    total_m/total_w
end
const mean_gsm = mean(gsm)
import Base.error
error(g::GSM) = begin
    regen_contrast(g, z_comp)
    total_w = 0.
    total_error = 0.
    C = inv(g.invC)
    #zs = 0.01:0.05:5.
    zs = [g.z_true]
    u1s = -3:0.1:3
    u2s = -3:0.1:3
    ps = zeros(length(u1s), length(u2s))
    for it = 1:length(zs) 
        z = zs[it]
        #S = inv(C + z*z*g.AᵀA/g.σₓ²)
        S = full(inv(PDMats.PDMat(g.invC + z*z*g.AᵀA/g.σₓ²)))
        m = (z/g.σₓ²) * S * g.Aᵀx
        #w = 1
        w = pdf(Normal(), z)*
            #Woodbury
            exp(dot(g.Aᵀx, (g.invC/(z*z) + g.AᵀA/g.σₓ²) \ g.Aᵀx)/(2*g.σₓ²^2)-
            #Generalisation of Sylvester matrix identity
                logdet(g.invC + z*z/g.σₓ²*g.AᵀA)/2)
        for i = 1:length(u1s), j=1:length(u2s)
            ps[i,j] += w*pdf(MvNormal(m[1:2], S[1:2, 1:2]), [u1s[i], u2s[j]])
        end
        total_w += w
        mean_error = m - mean_gsm
        total_error += w*(sum(mean_error.*mean_error) + sum(S))
    end
    #regen_contrast(g, 0.)
    (total_error/total_w, mat_to_df(ps/total_w, u1s, u2s))
    #mat_to_df(ps/total_w, u1s, u2s)
end
mat_to_df(ps::Matrix, u1s, u2s) = begin
    ps_ = Float64[]
    u1s_ = Float64[]
    u2s_ = Float64[]
    for i in 1:length(u1s), j in 1:length(u2s)
        push!(ps_, ps[i, j])
        push!(u1s_, u1s[i])
        push!(u2s_, u2s[j])
    end
    DataFrame(p=ps_, u1=u1s_, u2=u2s_)
end
const (mean_error, ps) = error(gsm)
#Pz() = zs[rand(Pz_)]
#posterior_error(g::GSM) = begin
#    println("start")
#    errors = Float64[]
#    n_zs = 10000
#    n_us = 1
#    mean_u = zeros(n)
#    C = inv(g.invC)
#    for i in 1:n_zs 
#        z = Pz()
#        Σ = inv(C + z*z*g.AᵀA/g.σₓ²)
#        μ = (z/g.σₓ²) * Σ * g.Aᵀx
#        #Σ = inv(C + z^2/g.σₓ²*g.AᵀA)
#        #μ = Σ*z/g.σₓ²*g.Aᵀx
#        dist = MvNormal(μ, Σ)
#        for j = 1:n_us
#            u = rand(dist)
#            error = u - mean_gsm
#            push!(errors, sum(error.*error))
#            mean_u += μ
#        end
#    end
#    println(length(errors))
#    println(mean_u/length(errors) - mean_gsm)
#    println("end")
#    mean(errors)
#end
#mean_error = posterior_error(gsm)


#Figure 1
mean_dkl(;Tms::Int = 1000, τh::Float64 = 10.) = begin
  interval = 1.
  s = Sampler(gsm, τh=τh)
  regen_contrast(s.dist, 0.)
  time_steps_per_interval = round(Int, interval/s.dt)
	total = 0
	totalu  = zeros(n)
	#totaluu = zeros(dim(s.dist)-1, dim(s.dist)-1)
  #Burn in 100 ms
  mean_distances = Float64[]
  for i = 1:round(Int, 100/s.dt)
    step(s)
  end
  regen_contrast(s.dist, z_comp)
  for t = 1:round(Int, Tms/interval)
    for j = 1:time_steps_per_interval
      step(s)
      total  += 1
      for k = 1:length(totalu)
        totalu[k] += s.u[k+1]
      end
      #totaluu += u*u'
    end
    m = totalu/total
    #s = totaluu/total - m*m'
    mdiff = m - mean_gsm
    push!(mean_distances, dot(mdiff, mdiff))
  end
  #println(totalu/total)
  mean_distances/mean_error
end
comp_time() = begin
    reps = 20#5#100
    runs_l = hcat([mean_dkl(τh=150.) for i = 1:reps]...)
    runs_h = hcat([mean_dkl(τh=10.) for i = 1:reps]...)
    mean_l = squeeze(mean(runs_l, 2), 2) 
    mean_h = squeeze(mean(runs_h, 2), 2) 
    var_l  = squeeze(var(runs_l, 2), 2) 
    var_h  = squeeze(var(runs_h, 2), 2) 
    error_threshold = 1.
    println(findfirst(x -> x < error_threshold, mean_l))
    println(findfirst(x -> x < error_threshold, mean_h))
    ts = collect(1:length(mean_l))
    df = vcat(DataFrame(t=ts, mean=mean_l, var=var_l, model="Langevin"),  
              DataFrame(t=ts, mean=mean_h, var=var_h, model="Hamiltonian"))
    writetable("comp_time", df)
    df
end


#Figure Langevin-Hamiltonian Comparison
langevin_hamiltonian_comparison() = begin
    regen_contrast(gsm, 1.)
    trace_ham = trace_uvs(Sampler(gsm), 100000, 1.)#, 0.1)
    trace_lan = trace_uvs(Sampler(gsm, τh=150.), 100000, 1.)#, 0.1)
    ts = collect(1:size(trace_ham[1], 1))
    df_ham = vcat(
                  DataFrame(t=trace_ham[1], uv="u", val=trace_ham[2][:, 2], u2=trace_ham[2][:, 3], L=trace_ham[4]),
      DataFrame(t=trace_ham[1], uv="v", val=trace_ham[3][:, 2]))
    df_lan = DataFrame(t=trace_lan[1], uv="u", val=trace_lan[2][:, 2], u2=trace_lan[2][:, 3], L=trace_lan[4])
    writetable("hamiltonian", df_ham[0 .== mod(collect(1:size(df_ham,1)), 10), :] )
    writetable("langevin", df_lan[0 .== mod(collect(1:size(df_lan,1)), 10), :])
    writetable("psn", ps)
    #writecsv("hamiltonian", trace_ham[1][:, 2:3])
    #writecsv("langevin", trace_lan[:, 2:3])
end

#Figure: EI-balance
ei_balance() = begin
    #gsm = GSM_contrast(100)
    regen_contrast(gsm, 0.)
    (us, vs) = trace_vs(Sampler(gsm), 4000, 1.)
    (u, v) = (mean(us, 2)[4001:end], mean(vs, 2)[4001:end])
    writetable("ei_balance", DataFrame(t=1:4000, u=u, v=v), separator=',')
end

ei_balance_stationary(iter=20) = begin
    ub = zeros(iter) 
    vb = zeros(iter) 
    for i = 1:iter
        (us, vs) = trace_vs(Sampler(GSM()), 1000, 1.)
        ub[i] = mean(us[1001:end, :])
        vb[i] = mean(vs[1001:end, :])
    end
    writetable("ei_balance_stationary", DataFrame(u=ub, v=vb))
end
    
  


#Figure 2
traces(us::Vector{Vector{Matrix{Float64}}}, zs::Vector) = begin
  df = DataFrame(t=Int[], u=Float64[], run=Int[], z=Float64[])
  flat_us = vcat(us...)
  flat_zs = repeat(zs, inner=[length(us[1])])
  halfway = round(Int, size(flat_us[1], 1)/2)
  half_interval = 1000
  for i = 1:length(flat_us)
    u = flat_us[i][(halfway-half_interval+1):(halfway+half_interval), 2]
    ndf = DataFrame(t=collect(1:length(u))-length(u)/2, u=u, run=i, z=flat_zs[i])
    df = vcat(df, ndf)
  end
  df
end

rates(us::Matrix{Float64}) = begin
  halfway = round(Int, size(us, 1)/2)
  half_interval = 400
  #squeeze(mean((us.>0) .* us, 2), 2)[(halfway-half_interval):(halfway+half_interval)]
  squeeze(mean(us .* us, 2), 2)[(halfway-half_interval):(halfway+half_interval)]
end

rates(us::Vector{Vector{Matrix{Float64}}}, zs::Vector{Float64}) = begin
  df = DataFrame(t=Int[], rate=Float64[], z=Float64[])
  for i = 1:length(us)
    rs = mean(map(rates, us[i]))
    ndf = DataFrame(t=collect(1:length(rs))-length(rs)/2, rate=rs, z=zs[i])
    df = vcat(df, ndf)
  end
  df
end

#Use Tms=10000, reps=10
ffts(us::Matrix{Float64}) = begin
  m_us = squeeze(mean(us[round(Int, size(us, 1)*0.7):end, 2:end], 2), 2)
  abs2(fft(m_us))
end
ffts(us::Vector{Matrix{Float64}}) =
  exp(mean(map(u -> log(ffts(u)), us)))
  #mean(map(u -> ffts(u), us))
ffts(us::Vector{Vector{Matrix{Float64}}}, zs::Vector) = begin
  df = DataFrame(freq=Int[], power=Float64[], z=Float64[])
  for i = 1:length(us)
    power = ffts(us[i])
    ndf = DataFrame(freq=collect(1:length(power))/length(power)*1000, power=power, z=zs[i])
    df = vcat(df, ndf)
  end
  df
end


mean_us() =
  hcat(map(u -> mean(u, 2), flat_us)...)
  
t_ffts(uss::Vector{Vector{Matrix{Float64}}}, zs::Vector{Float64}) =
  vcat(map(t_ffts, uss, zs)...)

t_ffts(us::Vector{Matrix{Float64}}, z::Float64) = begin
  T = 100
  width = 100
  l = size(us[1], 1)
  mean_us = map(u -> squeeze(mean(u[:, 2:end], 2), 2), us)
  result = zeros(l, T)
  ts = linspace(1+2width, l-2width, T)
  power = Float64[]
  freq  = Float64[]
  t     = Float64[]
  for i = 1:T
    kernel = exp(-((1:l)-ts[i]).^2/(2width^2))
    append!(freq, collect(1:l)/l*1000)
    append!(power, exp(mean(map(u -> log(abs(fft(u.*kernel))), mean_us))))
    append!(t, fill(ts[i], l))
  end
  DataFrame(freq=freq, power=power, t=t-round(Int, l/2), z=z)
end

si_fig(t::Int) = begin
  uzs = traces(zs=[0.5, 1., 2.], reps=10, τl=t)
  writetable("rate-tl=$t", rates(uzs...))
  writetable("fft-tl=$t", ffts(uzs...))
  uzs=traces(reps=4, different_models=false, τl=t)
  writetable("trace-tl=$t", traces(uzs...))
end
si_fig(ts::Vector{Int}) = map(si_fig, ts)

#si_fig_off_diag() = begin
#  uzs = traces(zs=[0.5, 1., 2.], reps=10)
#  writetable("rate-off-diag", rates(uzs...))
#  writetable("fft-off-diag", ffts(uzs...))
#  uzs=traces(reps=4, different_models=false)
#  writetable("trace-off-diag", traces(uzs...))
#end
#
vars(uzs) = begin
  uzs = uzs[1][1]
  us  = map(u -> u[:, 2:end], uzs)
  reps = length(us)
  Ers  = sum(us)/reps
  Ers2 = sum(map(r -> r.*r, us))/reps
  Varr = Ers2 - Ers.*Ers
  m = mean(sqrt(Varr), 2)
  halfway = round(Int, length(m)/2)
  half_interval=400
  m = m[(halfway-half_interval):(halfway+half_interval)]
  #DataFrame({[-half_interval:half_interval], m}, ["t", "m"])
  DataFrame(Any[[-half_interval:half_interval], m], ["t", "m"])
end








#Figure 3
contrast_fig(uzs_diff=traces(zs=[0.0, 0.5, 1., 2.], reps=20), #was 20
        uzs_same=traces(zs=[0.0, 0.5, 1., 2.], reps=4, different_models=false)) = begin
  writetable("contrast-fft", ffts(uzs_diff...), separator=',')
  writetable("contrast-tfft", t_ffts(uzs_diff...), separator=',')
  writetable("contrast-rate", rates(uzs_diff...), separator=',')
  writetable("contrast-traces", traces(uzs_same...), separator=',')
end

contrast_rate(uzs_diff=traces(zs=[0.5, 1., 2.], reps=1, τh=150.)) = 
  writetable("contrast-rate-langevin", rates(uzs_diff...), separator=',')
